@@ -2,10 +2,13 @@ import logging
 try:
     from PyQt5.QtCore import *
     from PyQt5.QtWebEngineWidgets import *
-    from PyQt5.QtWidgets import QApplication, QFileDialog, QMessageBox, QPlainTextEdit, QDialog, QSplashScreen, QProgressDialog, QMainWindow, QToolBar, QAction, QStyle
+    from PyQt5.QtWidgets import QApplication, QFileDialog, QMessageBox, QPlainTextEdit, QDialog, QSplashScreen, \
+        QMainWindow, QAction, QStyle, \
+        QDialogButtonBox, QFormLayout, \
+        QLabel, QListView, QPushButton, QLineEdit
     from PyQt5.QtWebChannel import QWebChannel
     from PyQt5 import QtGui 
-    from PyQt5.QtGui import QDesktopServices
+    from PyQt5.QtGui import QDesktopServices, QStandardItem, QStandardItemModel 
 
 except ImportError:
     # dependency missing, issue a warning
@@ -14,6 +17,7 @@ except ImportError:
     sys.exit()
 
 from optparse import OptionParser
+from functools import partial
 
 import pathlib
 from pathlib import Path
@@ -30,6 +34,8 @@ import urllib.parse
 import urllib.request
 import os
 import json
+import random
+import string
 
 # Don't remove this line.  The idna encoding
 # is used by getaddrinfo when dealing with unicode hostnames,
@@ -87,11 +93,84 @@ class textWindow(QDialog):
 #DEBUG_URL = 'http://127.0.0.1:%s' % DEBUG_PORT
 #os.environ['QTWEBENGINE_REMOTE_DEBUGGING'] = DEBUG_PORT
 
+
+class SelectPluginWindow(QDialog):
+    def __init__(self, app, parent=None):
+        try:
+            super(SelectPluginWindow, self).__init__(parent=parent)
+            self.app = app
+            self.setWindowTitle("Select Plugins")
+            form = QFormLayout(self)
+            form.addRow(QLabel("Plugin site:"))
+            self.textbox = QLineEdit(self)
+            self.textbox.setText("https://tissuumaps.github.io/TissUUmaps/plugins/")
+            form.addRow(self.textbox)
+
+            # Create a button in the window
+            self.button = QPushButton('Get plugins from site', self)
+            self.button.move(20,80)
+            form.addRow(self.button)
+            
+            # connect button to function on_click
+            self.button.clicked.connect(self.getPlugins)
+            form.addRow(QLabel("Available plugins:"))
+            self.listView = QListView(self)
+            form.addRow(self.listView)
+            
+            buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal, self)
+            form.addRow(buttonBox)
+            buttonBox.accepted.connect(self.accept)
+            buttonBox.rejected.connect(self.reject)
+
+            self.getPlugins()
+        except:
+            import traceback
+            print (traceback.format_exc())
+
+    def getPlugins(self):
+        self.url = self.textbox.text()
+        try:
+            response = urllib.request.urlopen(self.url + "pluginList.json")
+            self.items = json.loads(response.read())
+            for plugin in self.items:
+                if plugin["py"].replace(".py","") in self.app.config["PLUGINS"]:
+                    plugin["installed"] = True
+                else:
+                    plugin["installed"] = False
+            
+            model = QStandardItemModel(self.listView)
+            for item in self.items:
+                # create an item with a caption
+                standardItem = QStandardItem(item["name"])
+                standardItem.setCheckState(Qt.Checked if item["installed"] else Qt.Unchecked)
+                if not item["installed"]:
+                    standardItem.setCheckable(True)
+                standardItem.setEditable(False)
+
+                model.appendRow(standardItem)
+            self.listView.setModel(model)
+        except:
+            import traceback
+            try:
+                QMessageBox.warning(self, "Error", traceback.format_exc())
+            except:
+                print (traceback.format_exc())
+
+    def itemsSelected(self):
+        selected = []
+        model = self.listView.model()
+        i = 0
+        while model.item(i):
+            if model.item(i).checkState():
+                selected.append(self.items[i])
+            i += 1
+        return selected
+
 class MainWindow(QMainWindow):
     def __init__(self, qt_app, app, *args, **kwargs):
         super(MainWindow, self).__init__()
         self.resize(1400,1000)
-
+        self.app = app
         self.browser = webEngine(qt_app, app, self, *args)
         
         self.setCentralWidget(self.browser)
@@ -119,6 +198,10 @@ class MainWindow(QMainWindow):
             self.browser.page().runJavaScript("flask.standalone.saveProject();")
         _save.triggered.connect(trigger)
 
+        _close = QAction(self.style().standardIcon(QStyle.SP_DockWidgetCloseButton), "Close file",self)
+        file.addAction(_close)
+        _close.triggered.connect(self.browser.closeImage)
+
         file.addSeparator()
 
         _export = QAction(self.style().standardIcon(QStyle.SP_FileDialogListView), "Capture viewport",self)
@@ -141,13 +224,17 @@ class MainWindow(QMainWindow):
         _exit.triggered.connect(self.close)
 
         plugins = self.bar.addMenu("Plugins")
-        for pluginName in app.config["PLUGINS"]:
+        for pluginName in self.app.config["PLUGINS"]:
             _plugin = QAction(pluginName,self)
             plugins.addAction(_plugin)
-            def trigger():
-                print ("Plugin triggered:", pluginName)
-                self.browser.page().runJavaScript("pluginUtils.startPlugin(\""+pluginName+"\");");
-            _plugin.triggered.connect(trigger)
+
+            _plugin.triggered.connect(partial(self.triggerPlugin,pluginName))
+    
+        plugins.addSeparator()
+        _plugin = QAction("Add plugin",self)
+        plugins.addAction(_plugin)
+
+        _plugin.triggered.connect(self.addPlugin)
         
         about = self.bar.addMenu("About")
         _help = QAction(self.style().standardIcon(QStyle.SP_DialogHelpButton), "Help",self)
@@ -162,7 +249,6 @@ class MainWindow(QMainWindow):
                 folderPath = sys._MEIPASS
             else:
                 folderPath = os.path.dirname(pathlib.Path(__file__))
-            print (os.path.join(folderPath,"VERSION"), os.path.isfile(os.path.join(folderPath,"VERSION")))
             with open(os.path.join(folderPath,"VERSION"), "r") as fh:
                 version = fh.read()
             
@@ -170,6 +256,31 @@ class MainWindow(QMainWindow):
         _version.triggered.connect(trigger)
 
         self.showMaximized()
+    
+    def triggerPlugin(self, pName):
+        print ("Plugin triggered:", pName)
+        self.browser.page().runJavaScript("pluginUtils.startPlugin(\""+pName+"\");");
+
+    def addPlugin(self):
+        print ("Adding plugins")
+        try:
+            dial = SelectPluginWindow(self.app, self)
+            if dial.exec_() == QDialog.Accepted:
+                changed = False
+                for plugin in dial.itemsSelected():
+                    if not plugin["py"].replace(".py","") in self.app.config["PLUGINS"]:
+                        changed = True
+                        for type in ["py","js"]:
+                            if type in plugin.keys():
+                                urlFile = dial.url + plugin[type]
+                                localFile = os.path.join(self.app.config["PLUGIN_FOLDER_USER"],plugin[type])
+                                os.makedirs(self.app.config["PLUGIN_FOLDER_USER"], exist_ok=True)
+                                urllib.request.urlretrieve(urlFile, localFile)
+                if changed:
+                    QMessageBox.warning(self, "Restart TissUUmaps", "The new plugins will only be available after restarting TissUUmaps.")
+        except:
+            import traceback
+            print (traceback.format_exc())
 class webEngine(QWebEngineView):
     def __init__(self, qt_app, app, mainWin, args):
         super().__init__()
@@ -229,11 +340,14 @@ class webEngine(QWebEngineView):
             for url in event.mimeData().urls():
                 links.append(str(url.toLocalFile()))
             for link in links:
-                #link = link.replace("\\","/").replace(self.app.basedir.replace("\\","/"), "")
-                #print ("link",link, self.app.basedir)
-                if(".tmap") in link:
+                filename, file_extension = os.path.splitext(link)
+                print ("file_extension", file_extension)
+                if(file_extension == ".tmap"):
                     self.openImagePath(link)
-                self.page().runJavaScript(f"flask.standalone.addLayer(\"{link}\");")
+                elif(file_extension == ".csv"):
+                    self.page().runJavaScript(f"flask.standalone.addCSV(\"{link}\");")
+                else:
+                    self.page().runJavaScript(f"flask.standalone.addLayer(\"{link}\");")
             #self.emit(SIGNAL("dropped"), links)
         else:
             event.ignore()
@@ -256,8 +370,13 @@ class webEngine(QWebEngineView):
                     time.sleep(0.1)
                 else:
                     return
-                os.startfile(os.path.normpath(path))
-            
+                
+                if hasattr(os, 'startfile'):
+                    os.startfile(os.path.normpath(path))
+                else:
+                    opener = "open" if sys.platform == "darwin" else "xdg-open"
+                    subprocess.call([opener, os.path.normpath(path)])
+
             threading.Thread(target=openImageThread,daemon=True).start()
         
     def run (self):
@@ -298,6 +417,10 @@ class webEngine(QWebEngineView):
         folderpath = QFileDialog.getOpenFileName(self, 'Select a File',self.lastdir)[0]
         self.openImagePath(folderpath)
 
+    def closeImage(self):
+        self.load(QUrl(self.location))
+        self.mainWin.setWindowTitle("TissUUmaps")
+
     @pyqtSlot(str)
     def exportToStatic(self, state):
         imgFiles = []
@@ -384,13 +507,18 @@ class webEngine(QWebEngineView):
 
     @pyqtSlot(str)
     def saveProject(self, state):
-        def addRelativePath(state, relativePath):
+        def getRel(previouspath, file, newpath):
+            completepath = os.path.dirname(os.path.join(previouspath, file))
+            relPath = os.path.relpath(completepath, newpath)
+            return os.path.join(relPath, os.path.basename(file)).replace("\\","/")
+
+        def addRelativePath(state, previouspath, newpath):
             def addRelativePath_aux (state, path):
                 if len(path) == 1:
                     if isinstance(state[path[0]], list):
-                        state[path[0]] = [relativePath + "/" + s for s in state[path[0]]]
+                        state[path[0]] = [getRel(previouspath, s, newpath) for s in state[path[0]]]
                     else:
-                        state[path[0]] = relativePath + "/" + state[path[0]]
+                        state[path[0]] = getRel(previouspath, state[path[0]], newpath)
                     return
                 if not path[0] in state.keys():
                     return
@@ -402,7 +530,6 @@ class webEngine(QWebEngineView):
                         addRelativePath_aux (state[path[0]], path[1:])
             
             try:
-                relativePath = relativePath.replace("\\","/")
                 paths = [
                     ["layers","tileSource"],
                     ["markerFiles","path"],
@@ -416,20 +543,19 @@ class webEngine(QWebEngineView):
                 print (traceback.format_exc())
                 
             return state
-
-        parsed_url = urlparse(self.url().toString())
-        previouspath = parse_qs(parsed_url.query)['path'][0]
-        previouspath = os.path.abspath(os.path.join(self.app.basedir, previouspath))
-        print (previouspath)
-        
         folderpath = QFileDialog.getSaveFileName(self, 'Save project as',self.lastdir)[0]
         if (not folderpath):
             return {}
-
-        relativePath = os.path.relpath(previouspath, os.path.dirname(folderpath))
-        state = addRelativePath(json.loads(state), relativePath)
+        try:
+            parsed_url = urlparse(self.url().toString())
+            previouspath = parse_qs(parsed_url.query)['path'][0]
+            previouspath = os.path.abspath(os.path.join(self.app.basedir, previouspath))
+            print (previouspath)
+        except:
+            previouspath = self.app.basedir
+        state = addRelativePath(json.loads(state), previouspath, os.path.dirname(folderpath))
         with open(folderpath, "w") as f:
-            json.dump(state, f)
+            json.dump(state, f, indent=4)
 
     def openImagePath (self, folderpath):
         print ("openImagePath",folderpath)
@@ -470,6 +596,44 @@ class webEngine(QWebEngineView):
         self.close()
         #sys.exit()
 
+    @pyqtSlot(str, str, result="QJsonObject")
+    def addCSV(self, path, folderpath):
+        print ("gui.addCSV", path, folderpath)
+        if (folderpath == ""):
+            folderpath = QFileDialog.getOpenFileName(self, 'Select a File')[0]
+        if not folderpath:
+            returnDict = {"dzi":None,"name":None}
+            return returnDict
+        parts = Path(folderpath).parts
+        if (self.app.basedir != parts[0]):
+            if (not self.app.basedir == "C:\mnt\data\shared"):
+                reply = QMessageBox.warning(self, "Error", "All files must be in the same drive.")
+                returnDict = {"markerFile":None}
+                return returnDict
+            else:
+                self.app.basedir = parts[0]
+        imgPath = os.path.join(*parts[1:])
+        
+        path = os.path.abspath(os.path.join(self.app.basedir, path))
+        imgPath = os.path.abspath(os.path.join(self.app.basedir, imgPath))
+        
+        relativePath = os.path.relpath(os.path.dirname(imgPath), path) 
+        if ".." in relativePath:
+            reply = QMessageBox.question(self, "Error", "Impossible to add files from a parent folder.")
+            if reply == QMessageBox.Yes:
+                self.openImagePath(folderpath)
+            returnDict = {"dzi":None,"name":None}
+            return returnDict
+        returnDict = {
+            "markerFile":{
+                        "name": os.path.basename(imgPath),
+                        "path": relativePath + "/" + os.path.basename(imgPath),
+                        "uid": ''.join(random.choice(string.ascii_uppercase) for _ in range(6))
+                    }
+        }
+        print ("returnDict", returnDict)
+        return returnDict
+    
     @pyqtSlot(str, str, result="QJsonObject")
     def addLayer(self, path, folderpath):
         if (folderpath == ""):
@@ -483,12 +647,12 @@ class webEngine(QWebEngineView):
             if (not self.app.basedir == "C:\mnt\data\shared"):
                 reply = QMessageBox.question(self, "Error", "All layers must be in the same drive. Would you like to open this image only?")
                 reply = reply == QMessageBox.Yes
+                if reply:
+                    self.openImagePath(folderpath)
+                returnDict = {"dzi":None,"name":None}
+                return returnDict
             else:
-                reply = True
-            if reply:
-                self.openImagePath(folderpath)
-            returnDict = {"dzi":None,"name":None}
-            return returnDict
+                self.app.basedir = parts[0]
         imgPath = os.path.join(*parts[1:])
         try:
             views._get_slide(imgPath)
