@@ -25,6 +25,7 @@ glUtils = {
     _useColorFromMarker: {},     // {uid: boolean, ...}
     _useColorFromColormap: {},   // {uid: boolean, ...}
     _useScaleFromMarker: {},     // {uid: boolean, ...}
+    _useOpacityFromMarker: {},   // {uid: boolean, ...}
     _usePiechartFromMarker: {},  // {uid: boolean, ...}
     _useShapeFromMarker: {},     // {uid: boolean, ...}
     _colorscaleName: {},         // {uid: colorscaleName, ...}
@@ -66,6 +67,7 @@ glUtils._markersVS = `
     attribute float a_index;
     attribute float a_scale;
     attribute float a_shape;
+    attribute float a_opacity;
 
     varying vec4 v_color;
     varying vec2 v_shapeOrigin;
@@ -128,7 +130,7 @@ glUtils._markersVS = `
         v_shapeSize = gl_PointSize;
 
         // Discard point here in vertex shader if marker is hidden
-        v_color.a = v_color.a > 0.0 ? u_markerOpacity : 0.0;
+        v_color.a = v_color.a > 0.0 ? a_opacity * u_markerOpacity : 0.0;
         if (v_color.a == 0.0) DISCARD_VERTEX;
     }
 `;
@@ -198,6 +200,7 @@ glUtils._pickingVS = `
     uniform vec2 u_pickingLocation;
     uniform float u_markerScale;
     uniform float u_globalMarkerScale;
+    uniform float u_markerOpacity;
     uniform bool u_usePiechartFromMarker;
     uniform bool u_useShapeFromMarker;
     uniform int u_op;
@@ -207,6 +210,7 @@ glUtils._pickingVS = `
     attribute vec4 a_position;
     attribute float a_index;
     attribute float a_scale;
+    attribute float a_opacity;
 
     varying vec4 v_color;
 
@@ -270,6 +274,9 @@ glUtils._pickingVS = `
             uv = (uv + shapeOrigin) * (1.0 / SHAPE_GRID_SIZE);
             if (texture2D(u_shapeAtlas, uv).r < 0.5) DISCARD_VERTEX;
 
+            // Also do a quick alpha-test to avoid picking non-visible markers
+            if (a_opacity * u_markerOpacity <= 0.0) DISCARD_VERTEX
+
             // Output marker index encoded as hexadecimal color
             v_color.rgb = hex_to_rgb(a_index + float(u_op));
         }
@@ -381,11 +388,18 @@ glUtils.loadMarkers = function(uid) {
 
     const scalePropertyName = dataUtils.data[uid]["_scale_col"];
     const useScaleFromMarker = dataUtils.data[uid]["_scale_col"] != null;
-
     const markerScaleFactor = dataUtils.data[uid]["_scale_factor"];
     
     const sectorsPropertyName = dataUtils.data[uid]["_pie_col"];
     const usePiechartFromMarker = dataUtils.data[uid]["_pie_col"] != null;
+    if (dataUtils.data[uid]["_pie_dict"] && sectorsPropertyName) {
+        glUtils._piechartPalette = JSON.parse(dataUtils.data[uid]["_pie_dict"])
+        if (typeof glUtils._piechartPalette === "object") {
+            glUtils._piechartPalette = sectorsPropertyName.split(";").map(function(sector) {
+                return glUtils._piechartPalette[sector];
+            })
+        }
+    }
     const piechartPalette = glUtils._piechartPalette;
     let numSectors = 1;
 
@@ -394,15 +408,18 @@ glUtils.loadMarkers = function(uid) {
     const numShapes = Object.keys(markerUtils._symbolStrings).length;
     let shapeIndex = 0;
     
-    const markerOpacity = dataUtils.data[uid]["_opacity"];
+    const opacityPropertyName = dataUtils.data[uid]["_opacity_col"];
+    const useOpacityFromMarker = dataUtils.data[uid]["_opacity_col"] != null;
+    const markerOpacityFactor = dataUtils.data[uid]["_opacity"];
 
     // Additional info about the vertex format
-    const NUM_COMPONENTS_PER_MARKER = 7;
+    const NUM_COMPONENTS_PER_MARKER = 8;
     const NUM_BYTES_PER_MARKER = NUM_COMPONENTS_PER_MARKER * 4;
     const POINT_OFFSET = numPoints * 0,
           INDEX_OFFSET = numPoints * 4,
           SCALE_OFFSET = numPoints * 5,
           SHAPE_OFFSET = numPoints * 6;
+          OPACITY_OFFSET = numPoints * 7;
 
     // Extract and upload vertex data for markers. For datasets with tens of of
     // millions of points, the vertex data can be quite large, so we upload the
@@ -418,6 +435,7 @@ glUtils.loadMarkers = function(uid) {
         let bytedata_index = new Float32Array(chunkSize * 1);
         let bytedata_scale = new Float32Array(chunkSize * 1);
         let bytedata_shape = new Float32Array(chunkSize * 1);
+        let bytedata_opacity = new Float32Array(chunkSize * 1);
 
         if (usePiechartFromMarker) {
             // For piecharts, we need to create one marker per piechart sector,
@@ -427,6 +445,7 @@ glUtils.loadMarkers = function(uid) {
             bytedata_index = new Float32Array(chunkSize * numSectors * 1);
             bytedata_scale = new Float32Array(chunkSize * numSectors * 1);
             bytedata_shape = new Float32Array(chunkSize * numSectors * 1);
+            bytedata_opacity = new Float32Array(chunkSize * numSectors * 1);
 
             for (let i = 0; i < chunkSize; ++i) {
                 const markerIndex = i + offset;
@@ -448,6 +467,7 @@ glUtils.loadMarkers = function(uid) {
                     bytedata_shape[k] =
                         Math.floor((j < numSectors - 1 ? piechartAngles[j + 1] : 0.0) * 4095.0) +
                         Math.floor(piechartAngles[j] * 4095.0) * 4096.0;
+                    bytedata_opacity[k] = useOpacityFromMarker ? markerData[opacityPropertyName][markerIndex] : 1.0;
                 }
             }
         } else {
@@ -476,6 +496,7 @@ glUtils.loadMarkers = function(uid) {
                                                                  : Number("0x" + hexColor.substring(1,7));
                 bytedata_index[i] = markerIndex;  // Store index needed for picking
                 bytedata_scale[i] = useScaleFromMarker ? markerData[scalePropertyName][markerIndex] : 1.0;
+                bytedata_opacity[i] = useOpacityFromMarker ? markerData[opacityPropertyName][markerIndex] : 1.0;
             }
         }
 
@@ -505,13 +526,14 @@ glUtils.loadMarkers = function(uid) {
         gl.bufferSubData(gl.ARRAY_BUFFER, (INDEX_OFFSET + offset * 1) * numSectors * 4, bytedata_index);
         gl.bufferSubData(gl.ARRAY_BUFFER, (SCALE_OFFSET + offset * 1) * numSectors * 4, bytedata_scale);
         gl.bufferSubData(gl.ARRAY_BUFFER, (SHAPE_OFFSET + offset * 1) * numSectors * 4, bytedata_shape);
+        gl.bufferSubData(gl.ARRAY_BUFFER, (OPACITY_OFFSET + offset * 1) * numSectors * 4, bytedata_opacity);
         gl.bindBuffer(gl.ARRAY_BUFFER, null);
     }
     console.timeEnd("Generate vertex data");
 
     // Update marker info and LUT + colormap textures
     glUtils._numPoints[uid] = numPoints * numSectors;
-    glUtils._markerOpacity[uid] = markerOpacity;
+    glUtils._markerOpacity[uid] = markerOpacityFactor;
     glUtils._markerScalarRange[uid] = scalarRange;
     glUtils._markerScalarPropertyName[uid] = scalarPropertyName;
     glUtils._markerScaleFactor[uid] = markerScaleFactor;
@@ -519,6 +541,7 @@ glUtils.loadMarkers = function(uid) {
     glUtils._useColorFromMarker[uid] = useColorFromMarker;
     glUtils._useColorFromColormap[uid] = useColorFromColormap;
     glUtils._useScaleFromMarker[uid] = useScaleFromMarker;
+    glUtils._useOpacityFromMarker[uid] = useOpacityFromMarker;
     glUtils._usePiechartFromMarker[uid] = usePiechartFromMarker;
     glUtils._useShapeFromMarker[uid] = useShapeFromMarker;
     if (useColorFromColormap) {
@@ -545,7 +568,9 @@ glUtils.deleteMarkers = function(uid) {
     delete glUtils._useColorFromMarker[uid];
     delete glUtils._useColorFromColormap[uid];
     delete glUtils._useScaleFromMarker[uid];
+    delete glUtils._useOpacityFromMarker[uid];
     delete glUtils._usePiechartFromMarker[uid];
+    delete glUtils._useShapeFromMarker[uid];
     delete glUtils._colorscaleName[uid];
     delete glUtils._colorscaleData[uid];
     delete glUtils._barcodeToLUTIndex[uid];
@@ -775,7 +800,7 @@ glUtils._updateColorbarCanvas = function() {
         if (propertyMax.length > 9) propertyMax = propertyRange[1].toExponential(5);
         // Get marker tab name to show together with property name
         const tabName = interfaceUtils.getElementById(uid + "_marker-tab-name").textContent;
-        let label = tabName + "." + propertyName;
+        let label = tabName.substring(0, 15) + "." + propertyName.substring(0, 15);
 
         // Draw annotations (with drop shadow)
         ctx.font = "16px Segoe UI";
@@ -828,6 +853,7 @@ glUtils._loadTextureFromImageURL = function(gl, src) {
         gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, image);
         gl.generateMipmap(gl.TEXTURE_2D);  // Requires power-of-two size images
         gl.bindTexture(gl.TEXTURE_2D, null);
+        glUtils.draw();  // Force redraw to avoid black shapes after context loss
     };
     image.src = src;
     return texture;
@@ -845,6 +871,7 @@ glUtils.drawColorPass = function(gl, viewportTransform, markerScaleAdjusted) {
     const INDEX = gl.getAttribLocation(program, "a_index");
     const SCALE = gl.getAttribLocation(program, "a_scale");
     const SECTOR = gl.getAttribLocation(program, "a_shape");
+    const OPACITY = gl.getAttribLocation(program, "a_opacity");
 
     gl.uniform2fv(gl.getUniformLocation(program, "u_imageSize"), glUtils._imageSize);
     gl.uniform4fv(gl.getUniformLocation(program, "u_viewportRect"), glUtils._viewportRect);
@@ -869,6 +896,8 @@ glUtils.drawColorPass = function(gl, viewportTransform, markerScaleAdjusted) {
             gl.enableVertexAttribArray(SECTOR);
             gl.vertexAttribPointer(SECTOR, 1, gl.FLOAT, false, 0, numPoints * 24);
         }
+        gl.enableVertexAttribArray(OPACITY);
+        gl.vertexAttribPointer(OPACITY, 1, gl.FLOAT, false, 0, numPoints * 28);
 
         gl.uniform1f(gl.getUniformLocation(program, "u_globalMarkerScale"), glUtils._globalMarkerScale * glUtils._markerScaleFactor[uid]);
         gl.uniform2fv(gl.getUniformLocation(program, "u_markerScalarRange"), glUtils._markerScalarRange[uid]);
@@ -917,6 +946,7 @@ glUtils.drawPickingPass = function(gl, viewportTransform, markerScaleAdjusted) {
     const POSITION = gl.getAttribLocation(program, "a_position");
     const INDEX = gl.getAttribLocation(program, "a_index");
     const SCALE = gl.getAttribLocation(program, "a_scale");
+    const OPACITY = gl.getAttribLocation(program, "a_opacity");
 
     gl.uniform2fv(gl.getUniformLocation(program, "u_imageSize"), glUtils._imageSize);
     gl.uniform4fv(gl.getUniformLocation(program, "u_viewportRect"), glUtils._viewportRect);
@@ -940,10 +970,13 @@ glUtils.drawPickingPass = function(gl, viewportTransform, markerScaleAdjusted) {
         gl.enableVertexAttribArray(SCALE);
         gl.vertexAttribPointer(SCALE, 1, gl.FLOAT, false, 0, numPoints * 20);
         // Note: sectors for piecharts are currently not used in the picking
+        gl.enableVertexAttribArray(OPACITY);
+        gl.vertexAttribPointer(OPACITY, 1, gl.FLOAT, false, 0, numPoints * 28);
 
         gl.uniform1f(gl.getUniformLocation(program, "u_globalMarkerScale"), glUtils._globalMarkerScale * glUtils._markerScaleFactor[uid]);
         gl.uniform1i(gl.getUniformLocation(program, "u_usePiechartFromMarker"), glUtils._usePiechartFromMarker[uid]);
         gl.uniform1i(gl.getUniformLocation(program, "u_useShapeFromMarker"), glUtils._useShapeFromMarker[uid]);
+        gl.uniform1f(gl.getUniformLocation(program, "u_markerOpacity"), glUtils._markerOpacity[uid]);
         gl.activeTexture(gl.TEXTURE0);
         gl.bindTexture(gl.TEXTURE_2D, glUtils._textures[uid + "_colorLUT"]);
         gl.uniform1i(gl.getUniformLocation(program, "u_colorLUT"), 0);
@@ -988,7 +1021,7 @@ glUtils.draw = function() {
     let markerScaleAdjusted = glUtils._markerScale;
     if (glUtils._useMarkerScaleFix) markerScaleAdjusted *= (gl.canvas.height / 900.0);
 
-    gl.clearColor(0.0, 0.0, 0.0, 0.0);
+    gl.clearColor(0.5, 0.5, 0.5, 0.0);
     gl.clear(gl.COLOR_BUFFER_BIT);
     gl.viewport(0, 0, gl.canvas.width, gl.canvas.height);
 
@@ -1026,12 +1059,7 @@ glUtils.pick = function(event) {
             if (glUtils._usePiechartFromMarker[uid]) {
                 div.innerHTML = markerUtils.makePiechartTable(markerData, markerIndex, piechartPropertyName);
             } else {
-                if (groupName !== undefined) {
-	                div.innerHTML = groupName;
-                }
-                else if (dataUtils.data[uid]["_cb_col"]) {
-                    div.innerHTML = markerData[dataUtils.data[uid]["_cb_col"]][markerIndex];
-                }
+                div.innerHTML = markerUtils.getMarkerTooltip(uid, markerIndex);
                 console.log("Marker clicked:",tabName, groupName, "index:", markerIndex);
             }
             div.classList.add("viewer-layer", "m-0", "p-1");
@@ -1081,11 +1109,35 @@ glUtils.updateMarkerScale = function() {
 }
 
 
+glUtils._restoreLostContext = function(event) {
+    console.log("Restoring WebGL objects after context loss");
+    let canvas = document.getElementById("gl_canvas");
+    const gl = canvas.getContext("webgl", glUtils._options);
+
+    // Restore shared WebGL objects
+    glUtils._programs["markers"] = glUtils._loadShaderProgram(gl, glUtils._markersVS, glUtils._markersFS);
+    glUtils._programs["picking"] = glUtils._loadShaderProgram(gl, glUtils._pickingVS, glUtils._pickingFS);
+    glUtils._textures["shapeAtlas"] = glUtils._loadTextureFromImageURL(gl, "misc/markershapes.png");
+
+    // Restore per-markers WebGL objects
+    for (let [uid, numPoints] of Object.entries(glUtils._numPoints)) {
+        delete glUtils._buffers[uid + "_markers"];
+        delete glUtils._textures[uid + "_colorLUT"];
+        delete glUtils._textures[uid + "_colorscale"];
+        glUtils.loadMarkers(uid);
+    }
+
+    glUtils.draw();  // Make sure markers are redrawn
+}
+
+
 glUtils.init = function() {
     if (glUtils._initialized) return;
 
     let canvas = document.getElementById("gl_canvas");
     if (!canvas) canvas = this._createMarkerWebGLCanvas();
+    canvas.addEventListener("webglcontextlost", function(e) { e.preventDefault(); }, false);
+    canvas.addEventListener("webglcontextrestored", glUtils._restoreLostContext, false);
     const gl = canvas.getContext("webgl", glUtils._options);
 
     // Place marker canvas under the OSD canvas. Doing this also enables proper
