@@ -131,7 +131,7 @@ class ImageConverter:
                     if minVal == maxVal:
                         minVal = 0
                         maxVal = 255
-                    if (imgVips.percent(0.01) < 0 or imgVips.percent(99.99) > 255):
+                    if (imgVips.percent(1) < 0 or imgVips.percent(99) > 255):
                         imgVips = (255.0 * (imgVips - minVal)) / (maxVal - minVal)
                         imgVips = (imgVips < 0).ifthenelse(0, imgVips)
                         imgVips = (imgVips > 255).ifthenelse(255, imgVips)
@@ -143,6 +143,7 @@ class ImageConverter:
                         tile_width=256,
                         tile_height=256,
                         compression='jpeg',
+                        Q=95,
                         properties=True
                     )
                 except:
@@ -426,10 +427,12 @@ def csvFile(completePath):
     completePath = os.path.join(app.basedir, completePath + ".csv")
     directory = os.path.dirname(completePath)
     filename = os.path.basename(completePath)
-    print ("completePath", completePath, directory, filename, os.path.isfile(completePath))
     if os.path.isfile(completePath):
+        # We can not gzip csv files with the PapaParse library
         return send_from_directory(directory, filename)
-        # Temporary fix for gz files without csv
+        
+        # We keep the old gzip code anyway:
+        # gz files have to be names cgz for some browser
         if os.path.isfile(completePath + ".gz"):
             os.rename(completePath + ".gz", completePath + ".cgz")
 
@@ -472,10 +475,11 @@ def jsonFile(completePath):
 @app.route("/<path:path>.dzi")
 @requires_auth
 def dzi(path):
+    completePath = os.path.join(app.basedir, path)
     # Check if a .dzi file exists, else use OpenSlide:
-    if os.path.isfile(path + ".dzi"):
-        directory = os.path.dirname(path)
-        filename = os.path.basename(path) + ".dzi"
+    if os.path.isfile(completePath + ".dzi"):
+        directory = os.path.dirname(completePath)
+        filename = os.path.basename(completePath) + ".dzi"
         return send_from_directory(directory, filename)
     slide = _get_slide(path)
     format = app.config["DEEPZOOM_FORMAT"]
@@ -506,9 +510,10 @@ def dzi_asso(path):
 
 @app.route("/<path:path>_files/<int:level>/<int:col>_<int:row>.<format>")
 def tile(path, level, col, row, format):
-    if os.path.isfile( f"{path}_files/{level}/{col}_{row}.{format}"):
-        directory = os.path.dirname(f"{path}_files/{level}/{col}_{row}.{format}")
-        filename = os.path.basename(f"{path}_files/{level}/{col}_{row}.{format}")
+    completePath = os.path.join(app.basedir, path)
+    if os.path.isfile( f"{completePath}_files/{level}/{col}_{row}.{format}"):
+        directory = os.path.dirname(f"{completePath}_files/{level}/{col}_{row}.{format}")
+        filename = os.path.basename(f"{completePath}_files/{level}/{col}_{row}.{format}")
         return send_from_directory(directory, filename)
     slide = _get_slide(path)
     format = format.lower()
@@ -560,24 +565,20 @@ def load_plugin(name):
 
 @app.route("/plugins/<path:pluginName>.js")
 def runPlugin(pluginName):
-    directory = app.config["PLUGIN_FOLDER"]
+    for directory in [app.config["PLUGIN_FOLDER_USER"],app.config["PLUGIN_FOLDER"]]:        
+        filename = pluginName + ".js"
+        completePath = os.path.abspath(os.path.join(directory, pluginName + ".js"))
+        directory = os.path.dirname(completePath)
+        filename = os.path.basename(completePath)
+        if os.path.isfile(completePath):
+            return send_from_directory(directory, filename)
     
-    filename = pluginName + ".js"
-    completePath = os.path.abspath(os.path.join(directory, pluginName + ".js"))
-    directory = os.path.dirname(completePath)
-    filename = os.path.basename(completePath)
-    if os.path.isfile(completePath):
-        return send_from_directory(directory, filename)
-    else:
-        logging.error(completePath, "is not an existing file.")
-        abort(404)
+    logging.error(completePath, "is not an existing file.")
+    abort(404)
 
 
 @app.route("/plugins/<path:pluginName>/<path:method>", methods=["GET", "POST"])
 def pluginJS(pluginName, method):
-    logging.info("runPlugin", pluginName, method)
-    logging.debug(request.method)
-
     pluginModule = load_plugin(pluginName)
     pluginInstance = pluginModule.Plugin(app)
     pluginMethod = getattr(pluginInstance, method)
