@@ -9,7 +9,7 @@
  * @version projectUtils 2.0
  * @classdesc The root namespace for projectUtils.
  */
- projectUtils = {
+var projectUtils = {
      _activeState:{},
      _hideCSVImport: false,
      _settings:[
@@ -45,7 +45,7 @@
  projectUtils.saveProject = function(urlProject) {
     interfaceUtils.prompt("Save project under the name:","NewProject")
     .then((filename) => {
-        state = projectUtils.getActiveProject();
+        var state = projectUtils.getActiveProject();
         state.filename = filename;
 
         var dataStr = "data:text/json;charset=utf-8," + encodeURIComponent(JSON.stringify(state, null, 4));
@@ -119,7 +119,6 @@ projectUtils.getActiveProject = function () {
         }
         var title = document.getElementById("generateButtonTitle").value
         var comment = document.getElementById("generateButtonComment").value
-        console.log(_exists);
         if (!_exists) {
             interfaceUtils.confirm("Warning, path doesn't seem accessible on the server.\n\nAre you sure you want to continue?")
             .then(function(_confirm) {
@@ -174,6 +173,15 @@ projectUtils.getActiveProject = function () {
  }
 
 
+projectUtils.updateMarkerButton = function(dataset) {
+    var data_obj = dataUtils.data[dataset];
+    var markerFile = projectUtils._activeState.markerFiles[data_obj["fromButton"]];
+    var headers = interfaceUtils._mGenUIFuncs.getTabDropDowns(dataset);
+    markerFile.expectedHeader = Object.assign({}, ...Object.keys(headers).map((k) => ({[k]: headers[k].value})));
+    var radios = interfaceUtils._mGenUIFuncs.getTabRadiosAndChecks(dataset);
+    markerFile.expectedRadios = Object.assign({}, ...Object.keys(radios).map((k) => ({[k]: radios[k].checked})));
+}
+
 projectUtils.makeButtonFromTabAux = function (dataset, csvFile, title, comment) {
     buttonsDict = {};
 
@@ -220,7 +228,6 @@ projectUtils.loadProjectFile = function() {
         // here we tell the reader what to do when it's done reading...
         reader.onload = readerEvent => {
             var content = readerEvent.target.result; // this is the content!
-            console.log( content );
             projectUtils.loadProject(JSON.parse(content));
         }
     }
@@ -265,7 +272,6 @@ projectUtils.loadProjectFileFromServer = function(path) {
                 </div>
             </div>
         </div>`;
-        console.log(div)
         document.body.appendChild(div);
     }
     
@@ -346,11 +352,17 @@ projectUtils.loadProjectFileFromServer = function(path) {
             if (tab.visible === false) {document.getElementById("title-tab-" + tab.name).style.display="none"}
         });
     }*/
-    if (state.regions) {
+
+    if (state.regions && Object.keys(state.regions).length > 0) {
         regionUtils.JSONValToRegions(state.regions);
     }
     if (state.regionFile) {
-        regionUtils.JSONToRegions(state.regionFile);
+        const queryString = window.location.search;
+        const urlParams = new URLSearchParams(queryString);
+        const path = urlParams.get('path')
+        if (path != null) {
+            regionUtils.JSONToRegions(path + "/" + state.regionFile);
+        }
     }
     projectUtils._activeState = state;
     tmapp.fixed_file = "";
@@ -358,7 +370,8 @@ projectUtils.loadProjectFileFromServer = function(path) {
         filterUtils._compositeMode = state.compositeMode;
     }
     if (state.markerFiles) {
-        state.markerFiles.forEach(function(markerFile) {
+        state.markerFiles.forEach(function(markerFile, buttonIndex) {
+            markerFile["fromButton"] = buttonIndex;
             if (markerFile.expectedCSV) {
                 projectUtils.convertOldMarkerFile(markerFile);
                 state.hideTabs = true;
@@ -373,12 +386,12 @@ projectUtils.loadProjectFileFromServer = function(path) {
     }
     if (state.regionFiles) {
         state.regionFiles.forEach(function(regionFile) {
-            interfaceUtils.createDownloadButtonRegions(
-                regionFile.title,
-                regionFile.path,
-                regionFile.comment,
-                regionFile.autoLoad
-            );
+            if( Object.prototype.toString.call( regionFile.path ) === '[object Array]' ) {
+                interfaceUtils.createDownloadDropdownRegions(regionFile);
+            }
+            else {
+                interfaceUtils.createDownloadButtonRegions(regionFile);
+            }
         });
     }
     if (state.filename) {
@@ -389,6 +402,25 @@ projectUtils.loadProjectFileFromServer = function(path) {
         document.getElementById("project_title").href = state.link;
         document.getElementById("project_title").target = "_blank";
     }
+    if (state.settings) {
+        projectUtils.applySettings(state.settings);
+    }
+    if (state.hideTabs) {
+        document.getElementById("level-1-tabs").classList.add("d-none");
+    }
+    if (state.menuButtons) {
+        state.menuButtons.forEach(function(menuButton, i) {
+            interfaceUtils.addMenuItem([menuButton.text], function(){ window.open(menuButton.url, '_self').focus();});
+        });
+    }
+    projectUtils.loadLayers(state);
+    
+    //tmapp[tmapp["object_prefix"] + "_viewer"].world.resetItems()
+}
+
+/**
+ * This method is used to load the TissUUmaps layers from state */
+ projectUtils.loadLayers = function(state) {
     tmapp.layers = [];
     subfolder = window.location.pathname.substring(0, window.location.pathname.lastIndexOf('/'));
     state.layers.forEach(function(layer) {
@@ -401,7 +433,6 @@ projectUtils.loadProjectFileFromServer = function(path) {
         filterUtils._filtersUsed = state.filters;
         $(".filterSelection").prop("checked",false);
         state.filters.forEach(function(filterused, i) {
-            console.log("#filterCheck_" + filterused, $("#filterCheck_" + filterused));
             $("#filterCheck_" + filterused).prop("checked",true);
         });
     }
@@ -417,17 +448,6 @@ projectUtils.loadProjectFileFromServer = function(path) {
         filterUtils._compositeMode = state.compositeMode;
         filterUtils.setCompositeOperation();
     }
-    if (state.settings) {
-        state.settings.forEach(function(setting, i) {
-            try{
-                window[setting.module][setting.function] = setting.value;
-            }
-            catch (err) {}
-        });
-    }
-    if (state.hideTabs) {
-        document.getElementById("level-1-tabs").classList.add("d-none");
-    }
     /*if (projectUtils._hideCSVImport) {
         document.getElementById("ISS_data_panel").style.display="none";
     }*/
@@ -440,7 +460,6 @@ projectUtils.loadProjectFileFromServer = function(path) {
         if (state.boundingBox) {
             setTimeout(function() {
                 tmapp[tmapp["object_prefix"] + "_viewer"].viewport.fitBounds(new OpenSeadragon.Rect(state.boundingBox.x, state.boundingBox.y, state.boundingBox.width, state.boundingBox.height), false);
-                console.log("Changing BBox")
             },1000);
         }
         if (state.compositeMode) {
@@ -449,7 +468,6 @@ projectUtils.loadProjectFileFromServer = function(path) {
         }
         if (state.layerOpacities && state.layerVisibilities) {
             tmapp.layers.forEach(function(layer, i) {
-                console.log("state.layerOpacities[i]",i,state.layerOpacities[i])
                 $("#opacity-layer-"+i).val(state.layerOpacities[i]);
                 if (state.layerVisibilities[i] != 0) {
                     $("#visible-layer-"+i).click();
@@ -457,8 +475,6 @@ projectUtils.loadProjectFileFromServer = function(path) {
             });
         }
     },300);
-    
-    //tmapp[tmapp["object_prefix"] + "_viewer"].world.resetItems()
 }
 
 projectUtils.convertOldMarkerFile = function(markerFile) {
@@ -499,8 +515,8 @@ projectUtils.convertOldMarkerFile = function(markerFile) {
         markerFile.expectedRadios.cb_gr_rand = false;
         markerFile.expectedRadios.cb_gr_key = true;
         for (setting of markerFile.settings) {
-            if (setting.module == "glUtils" && setting.function == "_globalMarkerScale")
-                markerFile.expectedHeader.scale_factor = setting.value;
+            //if (setting.module == "glUtils" && setting.function == "_globalMarkerScale")
+            //    markerFile.expectedHeader.scale_factor = setting.value;
             if (setting.module == "markerUtils" && setting.function == "_selectedShape"){
                 dictSymbol = {6:6}
                 if (dictSymbol[setting.value]) setting.value = dictSymbol[setting.value];
@@ -519,8 +535,17 @@ projectUtils.convertOldMarkerFile = function(markerFile) {
                 markerFile.expectedRadios.cb_gr_dict = true;
                 markerFile.expectedHeader.cb_gr_dict = JSON.stringify(setting.value);
             }
-            if (setting.module == "glUtils" && setting.function == "_markerOpacity")
+            if (setting.module == "HTMLElementUtils" && setting.function == "_colorsperiter") {
+                markerFile.expectedRadios.cb_gr = true;
+                markerFile.expectedRadios.cb_gr_rand = false;
+                markerFile.expectedRadios.cb_gr_key = false;
+                markerFile.expectedRadios.cb_gr_dict = true;
+                markerFile.expectedHeader.cb_gr_dict = JSON.stringify(setting.value);
+            }
+            if (setting.module == "glUtils" && setting.function == "_markerOpacity") {
                 markerFile.expectedHeader.opacity = setting.value;
+                setting.function = "_markerOpacityOld"
+            }
             if (setting.module == "HTMLElementUtils" && setting.function == "_colorsperbarcode") {
                 markerFile.expectedRadios.cb_gr = true;
                 markerFile.expectedRadios.cb_gr_rand = false;
@@ -532,7 +557,6 @@ projectUtils.convertOldMarkerFile = function(markerFile) {
     }
     delete markerFile.expectedCSV;
     markerFile["hideSettings"] = true;
-    console.log(markerFile);
 }
 
 /**
@@ -559,3 +583,45 @@ projectUtils.commonPath = function(strs) {
     prefix = prefix.substring(0, prefix.lastIndexOf('/')+1);
     return prefix
 }
+
+/** Applying settings */
+projectUtils.applySettings = function (settings) {
+    if (settings) {
+        settings.forEach(function(setting, i) {
+            if (window[setting.module]) {
+                if (typeof window[setting.module][setting.function]  === 'function') {
+                    window[setting.module][setting.function](setting.value);
+                }
+                else {
+                    window[setting.module][setting.function] = setting.value;
+                }
+            }
+        });
+    }
+}
+
+/** Adding marker legend in the upper left corner */
+projectUtils.addLegend = function (htmlContent) {
+    if (! htmlContent) {
+        if (document.getElementById("markerLegend")) {
+            document.getElementById("markerLegend").style.display= "none";
+        }
+        return;
+    }
+    var op = tmapp["object_prefix"];
+    if (document.getElementById("markerLegend") == undefined) {
+        var elt = document.createElement('div');
+        elt.className = "px-1 mx-1 viewer-layer"
+        elt.id = "markerLegend"
+        elt.style.zIndex = "13";
+        elt.style.left = "10px";
+        elt.style.top = "10px";
+        elt.style.padding = "5px";
+        elt.style.overflowY = "auto";
+        elt.style.maxHeight = "Calc(100vh - 245px)";
+        tmapp[tmapp["object_prefix"] + "_viewer"].addControl(elt,{anchor: OpenSeadragon.ControlAnchor.TOP_LEFT});
+    }
+    elt = document.getElementById("markerLegend");
+    elt.style.display="block";
+    elt.innerHTML = htmlContent;
+}
